@@ -4,7 +4,7 @@
  *              to build a moderately scalable and highly available server
  *              based on a cluster of servers, with more flexibility.
  *
- * Version:     $Id: misc.c,v 1.6 2002/05/11 08:23:38 wensong Exp $
+ * Version:     $Id: misc.c,v 1.7 2003/02/14 09:21:32 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -104,7 +104,6 @@ struct socket * tcp_vs_connect2dest(struct tcp_vs_dest *dest)
 	EnterFunction(5);
 
 	/* First create a socket */
-
 	error = sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sock);
 	if (error < 0)
 		TCP_VS_ERR("Error during creation of socket; terminating\n");
@@ -122,10 +121,66 @@ struct socket * tcp_vs_connect2dest(struct tcp_vs_dest *dest)
 	}
 
 	LeaveFunction(5);
-
 	return sock;
 }
 
+
+/*
+ * tcp_vs_xmit is to send bytes from the buffer to the socket.
+ *
+ * A positive return-value indicates the number of bytes sent, a negative
+ * value indicates an error-condition.
+ *
+ * Note: tcp_vs_xmit will xmit all the bytes or fail.
+ */
+int
+tcp_vs_xmit (struct socket *sock, const char *buffer,
+		  const size_t length, unsigned long flags)
+{
+	struct msghdr	msg;
+	mm_segment_t	oldfs;
+	struct iovec	iov;
+	int		nbytes;	
+	int		len = 0;
+	int		offset = 0;
+	int		ret = length;
+
+	EnterFunction(6);
+
+	assert (buffer != NULL);
+	
+	nbytes = length;
+
+	msg.msg_name		= 0;
+	msg.msg_namelen		= 0;
+	msg.msg_iov		= &iov;
+	msg.msg_iovlen		= 1;
+	msg.msg_control		= NULL;
+	msg.msg_controllen	= 0;
+	msg.msg_flags		= MSG_NOSIGNAL | flags;
+	
+	while (nbytes > 0) {
+		msg.msg_iov->iov_base = (void *)buffer + offset;
+		msg.msg_iov->iov_len  = nbytes;
+
+		oldfs = get_fs(); set_fs(KERNEL_DS);
+		len = sock_sendmsg(sock, &msg, nbytes);
+		set_fs(oldfs);
+
+		if (len < 0) {
+			ret = -1;
+			break;
+		}
+		
+		nbytes -= len;
+		offset += len;
+	}
+
+	LeaveFunction(6);
+	return ret;
+}
+
+
 /*
  * tcp_vs_sendbuffer is to send bytes from the buffer to the socket.
  *
