@@ -11,7 +11,7 @@
 
 #include <asm/types.h>          /* For __uXX types */
 
-#define TCP_VS_VERSION_CODE		0x00000B
+#define TCP_VS_VERSION_CODE		0x00000C
 #define NVERSION(version)                       \
 	(version >> 16) & 0xFF,                 \
 	(version >> 8) & 0xFF,                  \
@@ -53,11 +53,6 @@
 
 #define TCP_VS_TEMPLATE_TIMEOUT 15*HZ
 
-/* constants for http parsing */
-#define HTTP_MAX_REQLINE_SIZE		512  /* # bytes of a request line (pool)    */
-#define HTTP_MAX_STALINE_SIZE		512  /* # bytes of a status line (pool)     */
-#define HTTP_MAX_MIMLINE_SIZE		512  /* # bytes of a mime line (pool)       */
-
 
 struct tcp_vs_ident {
 	char		name[KTCPVS_IDENTNAME_MAXLEN];
