@@ -6,7 +6,7 @@
  *
  * tcp_vs_http.c: KTCPVS content-based scheduling module for HTTP service
  *
- * Version:     $Id: tcp_vs_http.c,v 1.9 2002/05/11 08:23:38 wensong Exp $
+ * Version:     $Id: tcp_vs_http.c,v 1.10 2002/05/26 14:05:16 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -176,7 +176,8 @@ int parse_http_header(const char *buffer, size_t buflen, http_request_t *req)
 
 	EnterFunction(5);
 
-	TCP_VS_DBG(5, "parsing request:\n--------------------\n");
+	TCP_VS_DBG(5, "parsing request:\n");
+	TCP_VS_DBG(5, "--------------------\n");
 	TCP_VS_DBG(5, "%s\n", buffer);
 	TCP_VS_DBG(5, "--------------------\n");
 
@@ -297,8 +298,8 @@ tcp_vs_http_matchrule(struct tcp_vs_service *svc, http_request_t *req)
  *    Parse the http request, select a server according to the
  *    request, and create a socket the server finally.
  */
-static struct socket *tcp_vs_http_schedule(struct tcp_vs_conn *conn,
-					   struct tcp_vs_service *svc)
+static int
+tcp_vs_http_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
 {
 	struct tcp_vs_dest *dest;
 	struct socket *csock, *dsock;
@@ -322,7 +323,10 @@ static struct socket *tcp_vs_http_schedule(struct tcp_vs_conn *conn,
 	len = tcp_vs_recvbuffer(csock, buffer, buflen-1, MSG_PEEK);
 	if (len < 0) {
 		TCP_VS_ERR_RL("error reading request from client\n");
-		return NULL;
+		return -1;
+	} else if (len == 0) {
+		/* some clients may connect and disconnect immediately */
+		return 0;
 	}
 
 	/* make it zero-terminated string */
@@ -332,14 +336,14 @@ static struct socket *tcp_vs_http_schedule(struct tcp_vs_conn *conn,
 		TCP_VS_ERR_RL("cannot parse http request\n");
 
 		/* should redirect it to a local port next time */
-		return NULL;
+		return -1;
 	}
 
 	/*  Head.RemoteHost.s_addr = sock->sk->daddr; */
 
 	dest = tcp_vs_http_matchrule(svc, &req);
 	if (!dest)
-		return NULL;
+		return -1;
 
 	TCP_VS_DBG(5, "HTTP: server %d.%d.%d.%d:%d "
 		   "conns %d refcnt %d weight %d\n",
@@ -347,21 +351,23 @@ static struct socket *tcp_vs_http_schedule(struct tcp_vs_conn *conn,
 		   atomic_read(&dest->conns),
 		   atomic_read (&dest->refcnt), dest->weight);
 
-	atomic_inc(&dest->conns);
-	conn->dest = dest;
 	dsock = tcp_vs_connect2dest(dest);
 	if (!dsock) {
 		TCP_VS_ERR_RL("The destination is not available\n");
-		return NULL;
+		return -1;
 	}
 
+	atomic_inc(&dest->conns);
+	conn->dest = dest;
+	conn->dsock = dsock;
+
 	if (tcp_vs_sendbuffer(dsock, buffer, len, 0) != len) {
 		TCP_VS_ERR_RL("Error HTTP sending buffer\n");
 	}
 
 	LeaveFunction(5);
 
-	return dsock;
+	return 0;
 }
 
 
