@@ -7,7 +7,7 @@
  * tcp_vs_phttp.c: KTCPVS content-based scheduling module for HTTP service
  *		   with persistent connection support
  *
- * Version:     $Id: tcp_vs_phttp.c,v 1.2 2003/02/22 04:51:57 wensong Exp $
+ * Version:     $Id: tcp_vs_phttp.c,v 1.7 2003/04/07 13:59:17 wensong Exp $
  *
  * Authors:     Wensong Zhang, <wensong@linuxvirtualserver.org>
  *              Hai Long, <david_lung@yahoo.com>
@@ -65,11 +65,15 @@ typedef struct methods {
 } methods_t;
 
 
-const char * http_protocol_strings[] = {
-	"HTTP/0.9", /* The position is crucial & magic! do not insert, only add */
-	"HTTP/1.0",
-	"HTTP/1.1",
-	NULL
+#define HTTP_VERSION_HEADER_LEN		5
+#define HTTP_VERSION_NUMBER_LEN		3
+#define HTTP_VERSION_MAX		3
+
+const char* http_version_header = "HTTP/";
+const char * http_version_number[HTTP_VERSION_MAX] = {
+	"0.9", /* The position is crucial & magic! do not insert, only add */
+	"1.0",
+	"1.1",
 };
 
 const methods_t http_methods [HTTP_M_MAX] = {
@@ -138,6 +142,26 @@ typedef struct http_response_s
 	http_mime_header_t mime;
 } http_response_t;
 
+enum {
+	TRANSFER_ENCODING = 0,
+	CONTENT_LENGTH,
+	CONNECTION,
+	CONTENT_TYPE,
+	MAX_PARSER_ID
+};
+
+#define MAX_MIME_HEADER_STRING_LEN	64
+
+typedef void (*HTTP_MIME_PARSER) (http_mime_header_t *mime, char* buffer);
+
+typedef struct {
+	HTTP_MIME_PARSER parser;
+	char mime_header_string[MAX_MIME_HEADER_STRING_LEN];
+	int  mime_header_string_len;
+	int enable;
+} http_mime_parse_t;
+
+static http_mime_parse_t http_mime_parse_table[MAX_PARSER_ID];
 
 typedef struct http_read_ctl_block_s
 {
@@ -408,10 +432,16 @@ static int parse_http_request_line (char *buffer, size_t len, http_request_t *re
 
 	/* get http version */
 	req->version_str	= skip_lws (pos + 1);
-	req->version	= HTTP_V_UNKNOWN;
-	for (i = 0; http_protocol_strings[i] != NULL; i++) {
-		if (strnicmp (req->version_str, http_protocol_strings[i],
-			      strlen (http_protocol_strings[i])) == 0) {
+	if (strnicmp (req->version_str,
+		http_version_header, HTTP_VERSION_HEADER_LEN) != 0) {
+		goto exit;
+	}
+
+	req->version		= HTTP_V_UNKNOWN;
+	req->version_str	+= HTTP_VERSION_HEADER_LEN;
+	for (i = HTTP_VERSION_MAX - 1; i >= 0; i--) {
+		if (strncmp (req->version_str,
+			http_version_number[i], HTTP_VERSION_NUMBER_LEN) == 0) {
 			req->version = i + HTTP_V_0_9;
 			break;
 		}
@@ -421,7 +451,7 @@ static int parse_http_request_line (char *buffer, size_t len, http_request_t *re
 		goto exit;
 	}
 
-	TCP_VS_DBG(6, "HTTP VERSION: %s\n", http_protocol_strings[i]);
+	TCP_VS_DBG(6, "HTTP VERSION: %s\n", http_version_number[i]);
 	ret = PARSE_OK;
   exit:
 	buffer[len] = c;	/* restore string */
@@ -439,18 +469,20 @@ static int parse_http_request_line (char *buffer, size_t len, http_request_t *re
 *   accept any amount of SP or HT characters between fields, even though
 *   only a single SP is required.
 *
+*  Note: parse_http_status_line will modify the buffer, and will not restore it for
+*	   performance reason. so, I xmit the buffer before parse it.
+*
 */
 static int parse_http_status_line (char *buffer, size_t len, http_response_t *resp)
 {
-	char *pos, c;
+	char *pos;
 	int i, ret = PARSE_ERROR;
 
 	EnterFunction(5);
 
-	assert(buffer != NULL)
+	assert(buffer != NULL);
 
 	/* terminate string */
-	c = buffer[len];
 	buffer[len] = '\0';
 
 	TCP_VS_DBG(5, "parsing response:\n");
@@ -465,25 +497,30 @@ static int parse_http_status_line (char *buffer, size_t len, http_response_t *re
 
 	pos = skip_lws (buffer);
 
-	resp->version = HTTP_V_UNKNOWN;
-	for (i = 0; http_protocol_strings[i] != NULL; i++) {
-		if (strnicmp (pos, http_protocol_strings[i],
-			strlen (http_protocol_strings[i])) == 0) {
+	if (strnicmp (pos,
+		      http_version_header, HTTP_VERSION_HEADER_LEN) != 0) {
+		return ret;
+	}
+
+	pos		+= HTTP_VERSION_HEADER_LEN;
+	resp->version	= HTTP_V_UNKNOWN;
+	for (i = HTTP_VERSION_MAX - 1; i >= 0; i--) {
+		if (strnicmp (pos,
+			      http_version_number[i], HTTP_VERSION_NUMBER_LEN) == 0) {
 			resp->version = i + HTTP_V_0_9;
 			break;
 		}
 	}
 
+	pos += HTTP_VERSION_NUMBER_LEN;
 	if (resp->version != HTTP_V_UNKNOWN) {
 		/* get the status code */
-		pos += strlen (http_protocol_strings[i]) + 1;
 		resp->status_code = strtol (pos, NULL, 10);
-		assert(resp->status_code >= 100)
+		assert(resp->status_code >= 100);
 		TCP_VS_DBG(6, "Status Code: %d\n", resp->status_code);
 		ret = PARSE_OK;
 	}
 
-	buffer[len] = c;	/* restore string */
 	LeaveFunction(5);
 	return ret;
 }
@@ -497,7 +534,7 @@ tcp_vs_phttp_matchrule(struct tcp_vs_service *svc, http_request_t *req)
 	struct tcp_vs_dest *dest = NULL;
 	char *uri;
 
-	if (!(uri = kmalloc(req->uri_len+1, GFP_KERNEL))) {
+	if (!(uri = kmalloc(req->uri_len + 1, GFP_KERNEL))) {
 		TCP_VS_ERR("No memory!\n");
 		return NULL;
 	}
@@ -525,28 +562,31 @@ tcp_vs_phttp_matchrule(struct tcp_vs_service *svc, http_request_t *req)
 *
 * http_read_line - read a line from socket.
 *
-* Try to get the line from the remaining bytes. then read max_line_size bytes
-* from socket and get a line. (the line delimeter is "CRLF")
-* Return the len of the line (not including CRLF), or -1 if failed.
+*   At first, get the line from the remaining bytes. then read bytes without
+* move the buffer. finally, move the memory and read a full buffer to get a
+* line.
+*   Return the len of the line (not including CRLF), or -1 if failed.
 *
-*   Note: 1, http_read_line does not terminate the line with '\0', it still end
+*   Note:
+*	1, http_read_line does not terminate the line with '\0', it still end
 *   with CRLF.
-*	2, the acutul search length may bigger than max_line_size.
+*	2, If the line length is larger than the size of the buffer, it will
+*   failed.
+*
 */
-static int http_read_line (http_read_ctl_block_t *ctl_blk, int max_line_size)
+static int http_read_line (http_read_ctl_block_t *ctl_blk)
 {
 	char* buf;
-	int nbytes, i, offset, reads;
+	int nbytes, i, offset, reads, move;
 	int len = -1;
 
-	DECLARE_WAIT_QUEUE_HEAD(WQ);
+	DECLARE_WAITQUEUE(wait, current);
 
 	EnterFunction(5);
 
-	assert(max_line_size < ctl_blk->len);
 	assert(ctl_blk->remaining <= (ctl_blk->len - ctl_blk->offset));
 
-	ctl_blk->info = NULL;
+	ctl_blk->info		= NULL;
 	if (ctl_blk->remaining == 0) {
 		ctl_blk->offset = 0;
 	}
@@ -562,14 +602,10 @@ static int http_read_line (http_read_ctl_block_t *ctl_blk, int max_line_size)
 		}
 	}
 
-	/* overflow? */
-	if ((offset + max_line_size) > ctl_blk->len) {
-		memmove (ctl_blk->buffer, buf, ctl_blk->remaining);
-		ctl_blk->offset = 0;
-		buf = ctl_blk->buffer;
-	}
+	move = 0;
 
-	nbytes = max_line_size - ctl_blk->remaining;
+  get_a_line:
+	nbytes = ctl_blk->len - ctl_blk->offset - ctl_blk->remaining;
 
 	/* try to read a line from the socket */
 	while ((nbytes > 0) && (len < 0)) {
@@ -577,7 +613,7 @@ static int http_read_line (http_read_ctl_block_t *ctl_blk, int max_line_size)
 		if (ctl_blk->sock->sk->state != TCP_ESTABLISHED
 		    && ctl_blk->sock->sk->state != TCP_CLOSE_WAIT) {
 			if (len > 0)
-				break;
+				goto done;
 			else
 				goto exit;
 		}
@@ -589,8 +625,12 @@ static int http_read_line (http_read_ctl_block_t *ctl_blk, int max_line_size)
 					  ctl_blk->flag);
 
 		if (reads == 0) {
-			interruptible_sleep_on_timeout(&WQ, HZ);
 			TCP_VS_DBG(5, "Read 0 bytes while reading a line\n");
+			add_wait_queue(ctl_blk->sock->sk->sleep, &wait);
+			__set_current_state(TASK_INTERRUPTIBLE);
+			schedule_timeout(HZ);
+			__set_current_state(TASK_RUNNING);
+			remove_wait_queue(ctl_blk->sock->sk->sleep, &wait);
 			continue;
 		}
 
@@ -602,7 +642,7 @@ static int http_read_line (http_read_ctl_block_t *ctl_blk, int max_line_size)
 		ctl_blk->remaining += reads;
 
 		/* try to get a line from the remaing bytes */
-		for (; i < ctl_blk->remaining -1; i++) {
+		for (; i < ctl_blk->remaining - 1; i++) {
 			if ((buf[i] == CR) && (buf[i + 1] == LF)) {
 				len = i;
 				goto done;
@@ -611,6 +651,19 @@ static int http_read_line (http_read_ctl_block_t *ctl_blk, int max_line_size)
 		nbytes -= reads;
 	}
 
+	/* memmove and read again */
+	if ((len < 0) && (move == 0)) {
+		memmove (ctl_blk->buffer, buf, ctl_blk->remaining);
+		ctl_blk->offset = 0;
+		buf = ctl_blk->buffer;
+		move = 1;
+		goto get_a_line;
+	} else {
+		assert ((len < 0) && (move == 1));
+		TCP_VS_ERR("Buffer is too small while reading a line.\n");
+		goto exit;
+	}
+
   done:
 	ctl_blk->info		= buf;
 	ctl_blk->offset		+= len + 2;
@@ -625,6 +678,7 @@ static int http_read_line (http_read_ctl_block_t *ctl_blk, int max_line_size)
 }
 
 
+#if 0
 /****************************************************************************
 * http_line_unescape - convert escaped characters in buffer to ASCII
 *
@@ -645,7 +699,7 @@ static int http_read_line (http_read_ctl_block_t *ctl_blk, int max_line_size)
 static void http_line_unescape
 	(
 	char	*string,	/* escaped string to unescape */
-	int	len	/* length of the string */
+	int	len		/* length of the string */
 	)
 {
 	int	i = 0;
@@ -654,20 +708,20 @@ static void http_line_unescape
 
 	EnterFunction(5);
 
-	assert(string != NULL)
+	assert(string != NULL);
 
 	while (i < len) {
 		if (string[i] == '+') {
 			string[i] = ' ';	/* replace '+' by spaces */
 		}
 		if ((string[i] == '%') && (i < len-2)) {
-			if (isxdigit (string[i+1]) && isxdigit (string[i+2])) {
-				strncpy (buffer, &(string[i+1]), 2);
+			if (isxdigit (string[i + 1]) && isxdigit (string[i + 2])) {
+				strncpy (buffer, &(string[i + 1]), 2);
 				buffer[2] = 0;
 				c = (char) strtol (buffer, NULL, 16);
 				if (c != 0) {
 					memmove (&(string [i]),	/* move string 2 chars */
-						&(string[i+2]), 2);
+						&(string[i + 2]), 2);
 					string[i] = c;	/* replace % by new char */
 					len -= 2;
 				}
@@ -679,99 +733,196 @@ static void http_line_unescape
 	LeaveFunction(5);
 	return;
 }
+#endif
 
+/****************************************************************************
+*
+* register_mime_parser - register a http mime header parser
+*
+*/
+static void register_mime_parser
+	(
+	int parser_id,
+	HTTP_MIME_PARSER mime_parser,
+	char* mime_header_string
+	)
+{
+	assert (parser_id < MAX_PARSER_ID);
+	assert (mime_parser != NULL);
+	assert (mime_header_string != NULL);
+
+	strncpy (http_mime_parse_table[parser_id].mime_header_string,
+		mime_header_string, MAX_MIME_HEADER_STRING_LEN);
+	http_mime_parse_table[parser_id].mime_header_string_len = strlen (mime_header_string);
+	http_mime_parse_table[parser_id].parser = mime_parser;
+	http_mime_parse_table[parser_id].enable = 1;
+}
+
+/****************************************************************************
+*
+* http_mime_parse_enable - enable or disable a specified http mime header parser
+*
+*/
+static void http_mime_parse_enable (int parser_id, int enable)
+{
+	http_mime_parse_table[parser_id].enable = enable;
+}
+
+/****************************************************************************
+*
+* transfer_encoding_parser - http mime header parser for "Transfer-Encoding"
+*
+*/
+static void  transfer_encoding_parser (http_mime_header_t *mime, char* buffer)
+{
+	EnterFunction(6);
+
+	if (strnicmp (buffer, "identity", 8) != 0) {
+		mime->transfer_encoding = 1;
+		TCP_VS_DBG(6, "Transfer-Encoding: chunked\n");
+	}
+
+	LeaveFunction(6);
+	return;
+}
+
+
+/****************************************************************************
+*
+* content_length_parser - http mime header parser for "Content-Length"
+*
+*/
+static void content_length_parser (http_mime_header_t *mime, char* buffer)
+{
+	EnterFunction(6);
+
+	mime->content_length = strtol (buffer, NULL, 10);
+	TCP_VS_DBG(6, "Content-Length: %d\n", mime->content_length);
+
+	LeaveFunction(6);
+	return;
+}
+
+
+/****************************************************************************
+*
+* connection_parser - http mime header parser for "Connection"
+*
+*/
+static void connection_parser (http_mime_header_t *mime, char* buffer)
+{
+	EnterFunction(6);
+
+	if (strnicmp (buffer, "close", 5) == 0) {
+		mime->connection_close = 1;
+		TCP_VS_DBG(5, "Connection: close\n");
+	}
+
+	LeaveFunction(6);
+	return;
+
+}
+
+/****************************************************************************
+*
+* content_type_parser - http mime header parser for "Content-type"
+*
+* Note: buffer should be a NULL terminated string.
+*/
+static void content_type_parser (http_mime_header_t *mime, char* buffer)
+{
+	int sep_len;
+	char *pos;
+
+	EnterFunction(6);
+
+	if (strnicmp (buffer, "multipart/byteranges", 20) == 0){
+		TCP_VS_DBG(6, "multipart/byteranges\n");
+		pos = buffer + 20 + 1; /* skip ';' */
+		pos = skip_lws (pos + 1);
+		if (strnicmp (pos, "boundary=", 9) != 0) {
+			goto exit;
+		}
+
+		/* the rest of this line is THIS_STRING_SEPARATES */
+		pos += 9;
+		sep_len = strlen (pos);
+		if ((mime->sep = kmalloc (sep_len + 1, GFP_KERNEL)) == NULL) {
+			goto exit;
+		}
+
+		/* RFC 2046 [40] permits the boundary string to be quoted */
+		if (pos[0] == '"' || pos[0] == '\'') {
+			pos++;
+			sep_len--;
+		}
+		strncpy (mime->sep, pos, sep_len);
+		mime->sep[sep_len] = 0;
+		TCP_VS_DBG(5, "THIS_STRING_SEPARATES : %s\n", mime->sep);
+	}
+
+exit:
+	LeaveFunction(6);
+	return;
+}
+
+
+static void http_mime_parser_init (void)
+{
+	memset(http_mime_parse_table, 0, sizeof(http_mime_parse_table));
+	register_mime_parser (TRANSFER_ENCODING, transfer_encoding_parser, "Transfer-Encoding");
+	register_mime_parser (CONTENT_LENGTH, content_length_parser, "Content-Length");
+	register_mime_parser (CONNECTION, connection_parser, "Connection");
+	register_mime_parser (CONTENT_TYPE, content_type_parser, "Content-type");
+
+}
 
 /******************************************************************************
 * http_mime_parse - parse MIME line in a buffer
 *
 * This routine parses the MIME line in a buffer.
 *
-* NOTE: Some MIME headers (host, Referer) need be considered again, tbd.
+* NOTE: 1, Some MIME headers (host, Referer) need be considered again, tbd.
+*	2, http_mime_parse will modify the buffer, and will not restore it for
+*	   performance reason. so, I xmit the buffer before parse it.
 *
 */
 static int http_mime_parse
 	(
 	char			*buffer,
 	int			len,
-	http_mime_header_t	*mime,
-	int			partial
+	http_mime_header_t	*mime
 	)
 {
-	char	*pos, *buf, c;
-	int	l, ret = PARSE_ERROR;
+	char	*pos;
+	int	i, l, ret = PARSE_ERROR;
 
 	EnterFunction(5);
 
-	assert(buffer != NULL)
+	assert(buffer != NULL);
 
-		/* terminate string */
-		c = buffer[len];
+	/* terminate string */
 	buffer[len] = 0;
 
-	buffer = skip_lws (buffer);
-	if ((pos = strchr (buffer, ':')) == NULL) {
-		goto exit;
-	}
-
-	l = pos - buffer;
-
 	TCP_VS_DBG(5, "MIME Header: %s\n", buffer);
 
-	/*
-	 * Messages MUST NOT include both a Content-Length header field and
-	 * a non-identity transfer-coding. If the message does include a non-
-	 * identity transfer-coding, the Content-Length MUST be ignored.
-	 */
-	if (strnicmp (buffer, "Transfer-Encoding", l) == 0) { /* maybe strlen is better than 'l' */
-		pos = skip_lws (pos + 1);
-		if (strnicmp (pos, "identity", strlen ("identity")) != 0) {
-			mime->transfer_encoding = 1;
-			TCP_VS_DBG(5, "Transfer-Encoding: chunked\n");
-		}
+	buffer = skip_lws (buffer);
+	if ((pos = strchr (buffer, ':')) == NULL) {
+		return PARSE_ERROR;
 	}
-	else if (strnicmp (buffer, "Content-Length", l) == 0) {
-		mime->content_length = strtol (pos + 1, NULL, 10);
-		TCP_VS_DBG(5, "Content-Length: %d\n", mime->content_length);
-	} else if (strnicmp (buffer, "Connection", l) == 0) {
-		pos = skip_lws (pos + 1);
-		if (strnicmp (pos, "close", strlen ("close")) == 0) {
-			mime->connection_close = 1;
-			TCP_VS_DBG(5, "Connection: close\n");
-		}
-	}
-	else if (partial && (strnicmp (buffer, "Content-type", l) == 0)){
-		l = strlen ("multipart/byteranges");
-		pos = skip_lws (pos + 1);
-		if (strnicmp (pos, "multipart/byteranges", l)== 0){
-			TCP_VS_DBG(5, "multipart/byteranges\n");
-			pos += l + 1; /* skip ';' */
-			pos = skip_lws (pos + 1);
-			l = strlen ("boundary=");
-			if (strnicmp (pos, "boundary=", l) != 0) {
-				goto exit;
-			}
-			buf = pos + l;
 
-			/* the rest of this line is THIS_STRING_SEPARATES */
-			l = buffer + len - buf;
-			if ((mime->sep = kmalloc (l + 1, GFP_KERNEL)) == NULL) {
-				goto exit;
-			}
+	l = pos - buffer;
+	pos = skip_lws (pos + 1);
 
-			/* RFC 2046 [40] permits the boundary string to be quoted */
-			if (buf[0] == '"' || buf[0] == '\'') {
-				buf++;
-				l--;
-			}
-			strncpy (mime->sep, buf, l);
-			mime->sep[l] = 0;
-			TCP_VS_DBG(5, "THIS_STRING_SEPARATES : %s\n", mime->sep);
+	for (i = 0; i < MAX_PARSER_ID; i++) {
+		if (http_mime_parse_table[i].enable &&
+			(l == http_mime_parse_table[i].mime_header_string_len) &&
+			(strnicmp (http_mime_parse_table[i].mime_header_string, buffer, l) == 0)) {
+				 http_mime_parse_table[i].parser (mime, pos);
+				 break;
 		}
 	}
 
-	ret = PARSE_OK;
-  exit:
-	buffer[len] = c;	/* restore the string */
 	LeaveFunction(5);
 	return ret;
 }
@@ -789,18 +940,18 @@ static int relay_http_data
 	int nbytes, reads, w = 0;
 	int ret = -1;
 
-	DECLARE_WAIT_QUEUE_HEAD(WQ);
+	DECLARE_WAITQUEUE(wait, current);
 
 	EnterFunction(5);
 
-	assert(ctl_blk->remaining <= (ctl_blk->len - ctl_blk->offset))
-	assert(len > 0)
+	assert(ctl_blk->remaining <= (ctl_blk->len - ctl_blk->offset));
+	assert(len > 0);
 
 	/* if there is enough data in read buffer */
 	nbytes = len - ctl_blk->remaining;
 	if (nbytes <= 0) {
 		if (tcp_vs_xmit (dsock, ctl_blk->buffer + ctl_blk->offset,
-			len, MSG_MORE) < 0) {
+				 len, MSG_MORE) < 0) {
 			TCP_VS_ERR("Error in xmitting message body\n");
 			goto exit;
 		}
@@ -812,17 +963,21 @@ static int relay_http_data
 	/* xmit the remaining bytes */
 	if (ctl_blk->remaining > 0) {
 		if (tcp_vs_xmit (dsock, ctl_blk->buffer + ctl_blk->offset,
-			ctl_blk->remaining, MSG_MORE) < 0) {
-				TCP_VS_ERR("Error in xmitting remaining bytes\n");
-				goto exit;
+				 ctl_blk->remaining, MSG_MORE) < 0) {
+			TCP_VS_ERR("Error in xmitting remaining bytes\n");
+			goto exit;
 		}
 	}
 
 	do {
 		reads = tcp_vs_recvbuffer (ctl_blk->sock, ctl_blk->buffer, ctl_blk->len, ctl_blk->flag);
 		if (reads == 0) {
-			interruptible_sleep_on_timeout(&WQ, HZ);
-			TCP_VS_DBG (5, "Reads 0 bytes while relay\n");
+			TCP_VS_DBG(5, "Reads 0 bytes while relay\n");
+			add_wait_queue(ctl_blk->sock->sk->sleep, &wait);
+			__set_current_state(TASK_INTERRUPTIBLE);
+			schedule();
+			__set_current_state(TASK_RUNNING);
+			remove_wait_queue(ctl_blk->sock->sk->sleep, &wait);
 			continue;
 		}
 
@@ -844,11 +999,12 @@ static int relay_http_data
 	ctl_blk->offset		= w;
 	ctl_blk->remaining	= reads - w;
 
-	assert(ctl_blk->remaining >= 0)
-	assert(ctl_blk->offset  < ctl_blk->len)
-done:
+	assert(ctl_blk->remaining >= 0);
+	assert(ctl_blk->offset  < ctl_blk->len);
+
+  done:
 	ret = 0;
-exit:
+  exit:
 	LeaveFunction(5);
 	return ret;
 }
@@ -895,7 +1051,7 @@ static int relay_multiparts
 	char *buf, *pos;
 	char *sep = NULL;
 
-	DECLARE_WAIT_QUEUE_HEAD(WQ);
+	DECLARE_WAITQUEUE(wait, current);
 
 	EnterFunction(5);
 
@@ -927,8 +1083,13 @@ static int relay_multiparts
 	while (1) {
 		reads = tcp_vs_recvbuffer (ctl_blk->sock, ctl_blk->buffer + l, ctl_blk->len - l, 0);
 		if (reads == 0) {
-			interruptible_sleep_on_timeout(&WQ, HZ);
-			TCP_VS_DBG (5, "Reads 0 bytes while relaying multiparts\n");
+			TCP_VS_DBG(5, "Reads 0 bytes while relaying multiparts\n");
+			add_wait_queue(ctl_blk->sock->sk->sleep, &wait);
+			__set_current_state(TASK_INTERRUPTIBLE);
+			schedule();
+			__set_current_state(TASK_RUNNING);
+			remove_wait_queue(ctl_blk->sock->sk->sleep, &wait);
+			continue;
 		}
 
 		if (reads < 0) {
@@ -951,9 +1112,9 @@ static int relay_multiparts
 		memmove (ctl_blk->buffer, ctl_blk->buffer + len - l, l);
 	}
 
-done:
+  done:
 	ret = 0;
-exit:
+  exit:
 	if (sep) {
 		kfree (sep);
 	}
@@ -1033,7 +1194,7 @@ static int relay_http_message_body
 		 */
 		int len, chunk_size;
 		do {
-			len = http_read_line (ctl_blk, HTTP_MAX_MIMLINE_SIZE);
+			len = http_read_line (ctl_blk);
 			if (len < 0) {
 				TCP_VS_ERR("Error in reading chunk size from client\n");
 				goto exit;
@@ -1059,7 +1220,7 @@ static int relay_http_message_body
 
 		/* relay the trailer */
 		do {
-			len = http_read_line (ctl_blk, HTTP_MAX_MIMLINE_SIZE);
+			len = http_read_line (ctl_blk);
 			if (len < 0) {
 				TCP_VS_ERR("Error in reading trailer.\n");
 				goto exit;
@@ -1110,7 +1271,7 @@ http_get_response (struct socket *csock, struct socket* dsock,
 	}
 
 	/* read status line from server */
-	len = http_read_line (&read_ctl_blk, HTTP_MAX_STALINE_SIZE);
+	len = http_read_line (&read_ctl_blk);
 	if (len < 0) {
 		TCP_VS_ERR("Error in reading status line from server\n");
 		goto exit;
@@ -1129,9 +1290,11 @@ http_get_response (struct socket *csock, struct socket* dsock,
 		goto exit;
 	}
 
+	http_mime_parse_enable (CONTENT_TYPE, resp.status_code == 206);
+
 	/* parse MIME header */
 	do {
-		if ((len = http_read_line (&read_ctl_blk, HTTP_MAX_MIMLINE_SIZE)) < 0) {
+		if ((len = http_read_line (&read_ctl_blk)) < 0) {
 			goto exit;
 		}
 
@@ -1141,8 +1304,8 @@ http_get_response (struct socket *csock, struct socket* dsock,
 			goto exit;
 		}
 
-		http_line_unescape (read_ctl_blk.info, len);
-		http_mime_parse (read_ctl_blk.info, len, &resp.mime, resp.status_code == 206);
+		/* http_line_unescape (read_ctl_blk.info, len); */
+		http_mime_parse (read_ctl_blk.info, len, &resp.mime);
 	} while (len != 0); /* http header end with CRLF,CRLF */
 
 	*close = resp.mime.connection_close;
@@ -1237,7 +1400,6 @@ static int tcp_vs_phttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service
 	struct socket		*dsock;
 	server_conn_t		*sc;
 
-	/* DECLARE_WAIT_QUEUE_HEAD(WQ); */
 	DECLARE_WAITQUEUE(wait, current);
 
 	EnterFunction(5);
@@ -1251,11 +1413,6 @@ static int tcp_vs_phttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service
 	conn->dest	= NULL;
 	conn->dsock	= NULL;
 
-	if (read_ctl_blk.len < HTTP_MAX_REQLINE_SIZE) {
-		TCP_VS_ERR("Error: buffer is too small!\n");
-		goto out;
-	}
-
 	/* Do we have data ? */
 	while (skb_queue_empty(&(conn->csock->sk->receive_queue))) {
 		interruptible_sleep_on_timeout(&conn->csock->wait, HZ);
@@ -1290,10 +1447,9 @@ static int tcp_vs_phttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service
 				goto out;
 			}
 
-			/* interruptible_sleep_on_timeout(&WQ, HZ); */
 			add_wait_queue(read_ctl_blk.sock->sk->sleep, &wait);
 			__set_current_state(TASK_INTERRUPTIBLE);
-			schedule_timeout(5*HZ);
+			schedule_timeout(HZ);
 			__set_current_state(TASK_RUNNING);
 			remove_wait_queue(read_ctl_blk.sock->sk->sleep, &wait);
 			continue;
@@ -1304,7 +1460,7 @@ static int tcp_vs_phttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service
 		}
 
 		/* read request line from client socket */
-		len = http_read_line(&read_ctl_blk, HTTP_MAX_REQLINE_SIZE);
+		len = http_read_line(&read_ctl_blk);
 		if (len < 0) {
 			TCP_VS_ERR("Error reading request line from client\n");
 			ret = -2;
@@ -1400,9 +1556,11 @@ static int tcp_vs_phttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service
 			goto out_free;
 		}
 
+		http_mime_parse_enable (CONTENT_TYPE, 0);
+
 		/* Process MIME header */
 		do {
-			len = http_read_line (&read_ctl_blk, HTTP_MAX_MIMLINE_SIZE);
+			len = http_read_line (&read_ctl_blk);
 			if (len < 0) {
 				tcp_vs_srvconn_put(sc);
 				goto out;
@@ -1414,8 +1572,8 @@ static int tcp_vs_phttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service
 				goto out_free;
 			}
 
-			http_line_unescape (read_ctl_blk.info, len);
-			http_mime_parse (read_ctl_blk.info, len, &req.mime, 0);
+			/* http_line_unescape (read_ctl_blk.info, len); */
+			http_mime_parse (read_ctl_blk.info, len, &req.mime);
 		} while (len != 0); /* http header end with CRLF,CRLF */
 
 		if (relay_http_message_body(dsock, &read_ctl_blk, &req.mime) != 0) {
@@ -1465,6 +1623,7 @@ static struct tcp_vs_scheduler tcp_vs_phttp_scheduler =
 
 static int __init tcp_vs_phttp_init(void)
 {
+	http_mime_parser_init ();
 	INIT_LIST_HEAD(&tcp_vs_phttp_scheduler.n_list);
 	return register_tcp_vs_scheduler(&tcp_vs_phttp_scheduler);
 }
