@@ -1,7 +1,7 @@
 /*
  *      tcpvsadm - TCP Virtual Server ADMinistration program
  *
- *      Version: $Id: tcpvsadm.c,v 1.10 2003/02/22 04:51:59 wensong Exp $
+ *      Version: $Id: tcpvsadm.c,v 1.11 2003/02/27 02:54:40 wensong Exp $
  *
  *      Authors: Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -36,6 +36,7 @@
 #include <netinet/in.h>
 #include <sys/socket.h>
 #include <sys/types.h>
+#include <sys/wait.h>
 #include <sys/param.h>
 #include <arpa/inet.h>
 
@@ -68,6 +69,7 @@ static void fail(int err, char *msg, ...);
 static int list_service(struct tcp_vs_ident *id, unsigned int format);
 static int list_all(unsigned int format);
 static int load_configfile(char *cf);
+static int modprobe_ktcpvs(void);
 
 
 #define CMD_NONE		0x0000U
@@ -191,8 +193,11 @@ int main(int argc, char **argv)
 	struct tcp_vs_rule_u rule;
 	int result = 0;
 
-	if(tcpvs_init())
-		fail(2, "%s", tcpvs_strerror(errno));
+	if(tcpvs_init()) {
+		/* try to insmod the ip_vs module if ipvs_init failed */
+		if (modprobe_ktcpvs() || tcpvs_init())
+			fail(2, "%s", tcpvs_strerror(errno));
+	}
 
 	/* If no other arguement, list the ktcpvs service table */
 	if (argc == 1){
@@ -380,6 +385,28 @@ int main(int argc, char **argv)
 }
 
 
+static int modprobe_ktcpvs(void)
+{
+	char *argv[] = { "/sbin/modprobe", "-s", "-k", "--", "ktcpvs", NULL };
+	int child;
+	int status;
+	int rc;
+
+	if (!(child = fork())) {
+		execv(argv[0], argv);
+		exit(1);
+	}
+
+	rc = waitpid(child, &status, 0);
+
+	if (!WIFEXITED(status) || WEXITSTATUS(status)) {
+		return 1;
+	}
+
+	return 0;
+}
+
+
 static int load_configfile(char *cf)
 {
 	struct tcpvs_config config;
